/*
 * Copyright (C) 2013 David Suarez
 * See LICENSE for more information.
 */
#include <sys/mman.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include "libgpio.h"

// Direct access to the GPIO data and dir registers
volatile uint32_t *PORT0_21;
volatile uint32_t *DIR0_21;
volatile uint32_t *TOGGGLE0_21;
volatile uint32_t *PORT22_27;
volatile uint32_t *DIR22_27;

// Mapped memory for register access
struct mem_map
{
	unsigned long base_address;
	unsigned int size;

	void *mapped_memory;
	unsigned int mapped_offset;
	unsigned int mapped_size;
};

struct mem_map gpio_mem;

static void init_gpio_map()
{
	memset(&gpio_mem, 0, sizeof(gpio_mem));

    int npages = 0;
    int pagesize = getpagesize();

    gpio_mem.mapped_offset = GPIO_BASE_ADDRESS & (pagesize - 1);
    gpio_mem.base_address = GPIO_BASE_ADDRESS &  ~(pagesize - 1);

    npages += (GPIO_BASE_SIZE / pagesize);
    npages += 1;
    gpio_mem.mapped_size = npages * pagesize;
}

static void init_arduino_ports()
{
	uint32_t data = (uint32_t)gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_21_0_DATA_OFFSET;
	uint32_t dir = (uint32_t)gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_21_0_DIR_OFFSET;
	uint32_t toggle  = (uint32_t)gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_21_0_TOG_OFFSET;
	PORT0_21 = (uint32_t *)data;
	DIR0_21 = (uint32_t *)dir;
	TOGGGLE0_21 = (uint32_t *)toggle;

	data = (uint32_t)gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_27_22_DATA_OFFSET;
	dir = (uint32_t)gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_27_22_DIR_OFFSET;

	PORT22_27 = (uint32_t *)data;
	DIR22_27 = (uint32_t *)dir;	
}

// __ Public API ___________________________
void gpio_set_mode(uint32_t mode)
{
	uint32_t *addr = gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_MODE_OFFSET;
	uint32_t modeMask = 0x03;	
	uint32_t xh = (*addr & ~modeMask) | mode;
	*addr = xh;
}

uint32_t gpio_get_mode()
{
	uint32_t *addr = gpio_mem.mapped_memory + gpio_mem.mapped_offset + GPIO_MODE_OFFSET;
	return *addr;
}

void gpio_init()
{
	init_gpio_map();

	int iofd = open(MMAP_DEVICE, O_RDWR);
	if (iofd < 0)
		printf("open failed\n");

	gpio_mem.mapped_memory = mmap(NULL, gpio_mem.mapped_size,
			 PROT_READ|PROT_WRITE,
			 MAP_SHARED,
			 iofd, gpio_mem.base_address);

	if (gpio_mem.mapped_memory == MAP_FAILED)
		printf("mmap failed for %0lX\n", gpio_mem.base_address);

	init_arduino_ports();

	close(iofd);
}

void gpio_dispose()
{
	if (munmap(gpio_mem.mapped_memory, gpio_mem.mapped_size))
		printf("munmap failed for %lX\n", gpio_mem.base_address);
}

void gpio_toggle(uint32_t value)
{
	*TOGGGLE0_21 = value;
}

void gpio_setdir(uint32_t value)
{
	*DIR0_21 = value;
}

uint32_t gpio_getdir()
{
	return *DIR0_21;
}

inline void gpio_set(uint32_t value)
{
	*PORT0_21 = value;
}

inline uint32_t gpio_get()
{
	return *PORT0_21;
}

